#FILE: blast_hits_in_ucsc_genome_browser.pl
#AUTH: Paul Stothard (stothard@ualberta.ca)
#DATE: July 2, 2008
#VERS: 1.0

use warnings;
use strict;
use Data::Dumper;
use Getopt::Long;
use URI::Escape;
use File::Path;

#initialize options
my %options = (
    blast_file              => undef,
    genome_id               => undef,
    output_dir              => 3,
    url                     => undef,
    fs                      => '/',
    has_snp_position_column => 0,
    match_to_lowercase      => 0,
    verbose                 => 0,
    name_for_annotation     => 'BLAST results'
);

#get command line options
GetOptions(
    'b=s' => \$options{blast_file},
    'g=s' => \$options{genome_id},
    'd=s' => \$options{output_dir},
    'u=s' => \$options{url},
    'l'   => \$options{match_to_lowercase},
    'v'   => \$options{verbose},
    'n=s' => \$options{name_for_annotation}
);

#check for options
if (   !( defined( $options{blast_file} ) )
    or !( defined( $options{genome_id} ) )
    or !( defined( $options{output_dir} ) )
    or !( defined( $options{url} ) ) )
{
    die( print_usage() . "\n" );
}

#add trailing slash if missing from URL
if ( !( $options{url} =~ m/\/$/ ) ) {
    $options{url} = $options{url} . "/";
}

if ( !( -d $options{output_dir} ) ) {
    mkpath( [ $options{output_dir} ], { mode => 0775, verbose => 0 } )
        or die("Could not create directory '$options{output_dir}'.");
}

#get the blast results as an array of hashes
my @blast_results = @{ parse_blast_results( $options{blast_file} ) };

#change first character in match_id to lowercase if necessary
if ( $options{match_to_lowercase} ) {
    foreach (@blast_results) {
        if ( defined( $_->{match_id} ) ) {
            my $id = $_->{match_id};
            $_->{match_id} = lc( substr( $_->{match_id}, 0, 1 ) )
                . substr( $_->{match_id}, 1 );
        }
    }
}

#group HSPs by chromosome
my %annotations_by_chromosome = ();
foreach (@blast_results) {
    my $hsp = $_;
    if ( defined( $hsp->{match_id} ) ) {
        if ( !exists( $annotations_by_chromosome{ $hsp->{match_id} } ) ) {
            my @annotations = ();
            $annotations_by_chromosome{ $hsp->{match_id} } = \@annotations;
        }
        push( @{ $annotations_by_chromosome{ $hsp->{match_id} } }, $hsp );
    }
}

#write a separate annotation file for each chromosome
foreach ( keys(%annotations_by_chromosome) ) {
    my $chr = $_;
    open( my $ANNOTATION_FILE,
        '>', $options{output_dir} . $options{fs} . $chr . ".txt" )
        or die( "Cannot open file '"
            . $options{output_dir}
            . $options{fs}
            . $chr . ".txt"
            . "': $!" );

    #these 'browser lines' configure the overall display of the Genome Browser
    #window when the annotation file is uploaded. See
    #http://genome.ucsc.edu/goldenPath/help/hgTracksHelp.html#lines
    print {$ANNOTATION_FILE} "browser full refGene\n";
    print {$ANNOTATION_FILE}
        "browser pack all_mrna intronEst all_est multiz5way\n";

    #these 'track lines' define the display attributes for all lines in an
    #annotation data set. See
    #http://genome.ucsc.edu/goldenPath/help/hgTracksHelp.html#CustomTracks
    print {$ANNOTATION_FILE} "track name='$options{name_for_annotation}' visibility='2' db='$options{genome_id}'\n";

    my $count = 0;
    foreach ( @{ $annotations_by_chromosome{$chr} } ) {
        my $hsp = $_;
        $count++;

        my $start = $hsp->{s_start};
        my $end   = $hsp->{s_end};

#if there is a snp_position column, change the s_start and s_end to snp_position.
        if ( defined( $hsp->{snp_position} ) ) {
            $options{has_snp_position_column} = 1;
            $start                            = $hsp->{snp_position};
            $end                              = $hsp->{snp_position};
        }

        print {$ANNOTATION_FILE}
            "$hsp->{match_id}\tBLAST\t$hsp->{query_id}\t$start\t$end\t.\t$hsp->{hit_strand}\t.\t$hsp->{query_id}($count)\n";
    }

    close($ANNOTATION_FILE) or die("Cannot close file : $!");
}

#create an html file containing links to the UCSC Genome Browser, one per HSP
open( my $HTML_FILE, '>', $options{output_dir} . $options{fs} . "index.html" )
    or die( "Cannot open file '"
        . $options{output_dir}
        . $options{fs}
        . "index.html"
        . "': $!" );

print {$HTML_FILE} "<html><body><pre>\n";
if ( $options{has_snp_position_column} ) {
    if ( $options{verbose} ) {
        print {$HTML_FILE}
            "query\tsubject\t\%_identity\tq_start\tq_end\ts_start\ts_end\tstrand\tsnp_position\tlink\n";
    }
    else {
        print {$HTML_FILE} "query\tlink\n";
    }
}
else {
    if ( $options{verbose} ) {
        print {$HTML_FILE}
            "query\tsubject\t\%_identity\tq_start\tq_end\ts_start\ts_end\tstrand\tlink\n";
    }
    else {
        print {$HTML_FILE} "query\tlink\n";
    }
}

my $count = 0;

foreach (@blast_results) {
    $count++;
    my $hsp = $_;

    if ( !defined( $hsp->{match_id} ) ) {
        if ( $options{has_snp_position_column} ) {
            if ( $options{verbose} ) {
                print {$HTML_FILE}
                    "$hsp->{query_id}\t$hsp->{comment}\t-\t-\t-\t-\t-\t-\t-\t-\n";
            }
            else {
                print {$HTML_FILE} "$hsp->{query_id}\t-\n";
            }
        }
        else {
            if ( $options{verbose} ) {
                print {$HTML_FILE}
                    "$hsp->{query_id}\t$hsp->{comment}\t-\t-\t-\t-\t-\t-\t-\n";
            }
            else {
                print {$HTML_FILE} "$hsp->{query_id}\t-\n";
            }
        }
        next;
    }

#these 'gff lines' describe how each BLAST HSP relates to the hit chromosome:
#   1. seqname - The name of the sequence. Must be a chromosome or scaffold.
#   2. source - The program that generated this feature.
#   3. feature - The name of this type of feature. Some examples of standard feature types are "CDS", "start_codon", "stop_codon", and "exon".
#   4. start - The starting position of the feature in the sequence. The first base is numbered 1.
#   5. end - The ending position of the feature (inclusive).
#   6. score - A score between 0 and 1000. If the track line useScore attribute is set to 1 for this annotation data set, the score value will determine the level of gray in which this feature is displayed (higher numbers = darker gray). If there is no score value, enter ".".
#   7. strand - Valid entries include '+', '-', or '.' (for don't know/don't care).
#   8. frame - If the feature is a coding exon, frame should be a number between 0-2 that represents the reading frame of the first base. If the feature is not a coding exon, the value should be '.'.
#   9. group - All lines with the same group are linked together into a single item.

#the syntax for linking to the UCSC browser is:
#http://genome.ucsc.edu/cgi-bin/hgTracks?db=database_name&position=chr_position&hgt.customText=URL.

    my $url
        = "http://genome.ucsc.edu/cgi-bin/hgTracks?db="
        . uri_escape( $options{genome_id} )
        . "\&position="
        . uri_escape( $hsp->{match_id} ) . ":"
        . uri_escape( $hsp->{s_start} ) . "-"
        . uri_escape( $hsp->{s_end} )
        . "\&hgt.customText=$options{url}$hsp->{match_id}.txt";

    if ( $options{has_snp_position_column} ) {
        if ( $options{verbose} ) {
            print {$HTML_FILE}
                "$hsp->{query_id}\t$hsp->{match_id}\t$hsp->{'%_identity'}\t$hsp->{q_start}\t$hsp->{q_end}\t$hsp->{s_start}\t$hsp->{s_end}\t$hsp->{hit_strand}\t$hsp->{snp_position}\t<a href=\"$url\" target=\"ucsc\">view</a>\n";
        }
        else {
            print {$HTML_FILE}
                "$hsp->{query_id}\t<a href=\"$url\" target=\"ucsc\">view</a>\n";
        }
    }
    else {
        if ( $options{verbose} ) {
            print {$HTML_FILE}
                "$hsp->{query_id}\t$hsp->{match_id}\t$hsp->{'%_identity'}\t$hsp->{q_start}\t$hsp->{q_end}\t$hsp->{s_start}\t$hsp->{s_end}\t$hsp->{hit_strand}\t<a href=\"$url\" target=\"ucsc\">view</a>\n";
        }
        else {
            print {$HTML_FILE}
                "$hsp->{query_id}\t<a href=\"$url\" target=\"ucsc\">view</a>\n";
        }
    }
}

print {$HTML_FILE} "</pre></body></html>\n";

close($HTML_FILE) or die("Cannot close file : $!");

sub parse_blast_results {
    my $file = shift;

#the file can contain comments starting with'#'
#the file must have a line beginning with a 'query_id' indicating the column names:

    my $header_count = 0;
    my @column_names = ();
    my $columns_read = 0;

    open( my $INFILE, $file ) or die("Cannot open file '$file': $!");

    #obtain column names
    while ( my $line = <$INFILE> ) {
        $line =~ s/\cM|\n//g;
        $header_count++;
        if ( $line =~ m/^\#/ ) {
            next;
        }
        if ( $line =~ m/^query_id/ ) {
            $columns_read = 1;
            @column_names = @{ _split($line) };
            last;
        }
    }

    #read the remaining entries
    my @entries    = ();
    my $line_count = 0;
    while ( my $line = <$INFILE> ) {
        $line =~ s/\cM|\n//g;
        $line_count++;

        if ( $line =~ m/\S/ ) {
            my %entry  = ();
            my @values = @{ _split($line) };

            if ( scalar(@values) != scalar(@column_names) ) {

                #most likely not hit was obtained, i.e something like:
                #53_[T/C]_BTA-22554 no acceptable hits returned
                if ( scalar(@values) == 2 ) {
                    $entry{query_id} = $values[0];
                    $entry{comment}  = $values[1];
                    push( @entries, \%entry );
                }

                next;
            }

            for ( my $i = 0; $i < scalar(@column_names); $i++ ) {
                $entry{ $column_names[$i] } = $values[$i];
            }

     #depending on the blast search, the s_start may be larger than the s_end,
     #to indicate that the query aligns to the reverse strand of the hit.
            if ( $entry{s_start} > $entry{s_end} ) {
                ( $entry{s_start}, $entry{s_end} )
                    = ( $entry{s_end}, $entry{s_start} );
                $entry{hit_strand} = '-';
            }
            else {
                $entry{hit_strand} = '+';
            }

            push( @entries, \%entry );
        }
    }
    close($INFILE) or die("Cannot close file : $!");
    return \@entries;
}

sub _split {
    my $line   = shift;
    my @values = ();
    if ( $line =~ m/\t/ ) {
        @values = split( /\t/, $line );
    }
    else {
        @values = split( /\,/, $line );
    }
    foreach (@values) {
        $_ = clean_value($_);
    }
    return \@values;
}

sub clean_value {
    my $value = shift;
    if ( !defined($value) ) {
        return ".";
    }
    if ( $value =~ m/^\s*$/ ) {
        return ".";
    }
    $value =~ s/^\s+//g;
    $value =~ s/\s+$//g;
    $value =~ s/\"|\'//g;
    return $value;
}

sub print_usage {
    print <<BLOCK;
USAGE: perl blast_hits_in_ucsc_genome_browser.pl [-arguments]
 -b [FILE]      : the BLAST results file to process (Required).
 -g [STRING]    : genome / assembly ID from UCSC ('bosTau4' for
                  example) (Required).
 -d [DIRECTORY] : output directory (Required).
 -u [URL]       : final full URL of the output directory (Required).
 -l             : change first letter in match ID to lowercase ('ChrX'
                  to 'chrX' for example) (Optional).
 -v             : give verbose HTML output (Optional).
 -n [STRING]    : name for UCSC annotation track ('SNPs' for example)
                  (Optional).

perl blast_hits_in_ucsc_genome_browser.pl -b blast_results.txt \
-g bosTau4 -d output_dir -u 'http://my_site/output_dir'
BLOCK
}
